import json
from fastapi import Request
from blocks_genesis._auth.blocks_context import BlocksContext, BlocksContextManager
from blocks_genesis._database.db_context import DbContext
from blocks_genesis._lmt.activity import Activity
from blocks_genesis._tenant.tenant_service import get_tenant_service
from motor.motor_asyncio import AsyncIOMotorCollection


async def change_context(project_key: str, request: Request | None = None):
    # STEP 1: Read third-party context header (same as C#)
    if request:
        await apply_third_party_context(request)

    # After third-party context applied, read the current context
    context = BlocksContextManager.get_context()

    # Track actual tenant in OTel baggage
    Activity.set_current_property("baggage.ActualTenantId", context.tenant_id)

    # Skip if project_key invalid or same as current tenant
    if not project_key or project_key == context.tenant_id:
        return

    tenant_service = get_tenant_service()
    tenant = await tenant_service.get_tenant(project_key)

    # Check whether the user is in the shared project
    collection: AsyncIOMotorCollection = await  DbContext.get_provider().get_collection("ProjectPeoples")
    shared_project =  collection.find_one(
        {"UserId": context.user_id, "TenantId": project_key}
    )

    # Check root tenant flag
    is_root = (await tenant_service.get_tenant(context.tenant_id)).is_root_tenant

    # Same C# condition:
    # isRoot && (tenant.CreatedBy == userId || sharedProject exists)
    if is_root and (tenant.created_by == context.user_id or shared_project):
        BlocksContextManager.set_context(
            BlocksContextManager.create(
                tenant_id=project_key,
                roles=context.roles,
                user_id=context.user_id,
                is_authenticated=context.is_authenticated,
                request_uri=context.request_uri,
                organization_id=context.organization_id,
                expire_on=context.expire_on,
                email=context.email,
                permissions=context.permissions,
                user_name=context.user_name,
                phone_number=context.phone_number,
                display_name=context.display_name,
                oauth_token=context.oauth_token,
                actual_tenant_id=context.tenant_id,
            )
        )


async def apply_third_party_context(request: Request):
    
    third_party_header = getattr(request.state, "third_party_context_header", None)

    if not third_party_header:
        return

    try:
        context_dict = json.loads(third_party_header)
    except json.JSONDecodeError:
        return

    # Create new context object
    new_context = BlocksContextManager.create(
        tenant_id=context_dict.get("tenant_id", ""),
        roles=context_dict.get("roles", []),
        user_id=context_dict.get("user_id", ""),
        is_authenticated=context_dict.get("is_authenticated", False),
        request_uri=context_dict.get("request_uri", ""),
        organization_id=context_dict.get("organization_id", ""),
        expire_on=context_dict.get("expire_on"),
        email=context_dict.get("email", ""),
        permissions=context_dict.get("permissions", []),
        user_name=context_dict.get("user_name", ""),
        phone_number=context_dict.get("phone_number", ""),
        display_name=context_dict.get("display_name", ""),
        oauth_token=context_dict.get("oauth_token", ""),
        actual_tenant_id=context_dict.get("actual_tent_id", "")
    )

    BlocksContextManager.set_context(new_context)



from typing import Dict, Optional

from pydantic import BaseModel

class ConsumerMessage(BaseModel):
    consumer_name: str
    payload: Dict
    payload_type: str
    context: Optional[str] = None
    routing_key: str = ""




from datetime import datetime
from typing import List, Optional
from pydantic import BaseModel, Field

class BaseEntity(BaseModel):
    id: str = Field(alias="_id")
    created_date: datetime = Field(alias="CreatedDate", default_factory=datetime.now)
    last_updated_date: datetime = Field(alias="LastUpdatedDate", default_factory=datetime.now)
    created_by: Optional[str] = Field(alias="CreatedBy", default=None)
    language: Optional[str] = Field(alias="Language", default=None)
    last_updated_by: Optional[str] = Field(alias="LastUpdatedBy", default=None)
    organization_ids: List[str] = Field(alias="OrganizationIds", default_factory=list)
    tags: List[str] = Field(alias="Tags", default_factory=list)


import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from fastapi import FastAPI
from blocks_genesis._core import api

@pytest.mark.asyncio
@patch('blocks_genesis._core.api.SecretLoader')
@patch('blocks_genesis._core.api.configure_logger')
@patch('blocks_genesis._core.api.configure_tracing')
@patch('blocks_genesis._core.api.CacheProvider')
@patch('blocks_genesis._core.api.RedisClient')
@patch('blocks_genesis._core.api.initialize_tenant_service', new_callable=AsyncMock)
@patch('blocks_genesis._core.api.DbContext')
@patch('blocks_genesis._core.api.MongoDbContextProvider')
@patch('blocks_genesis._core.api.AzureMessageClient')
@patch('blocks_genesis._core.api.MessageConfiguration')
async def test_configure_lifespan(mock_msg_config, mock_client, mock_mongo, mock_db, mock_init_tenant, mock_redis, mock_cache, mock_tracing, mock_logger, mock_secret_loader):
    msg_config = MagicMock()
    await api.configure_lifespan('svc', msg_config)

@pytest.mark.asyncio
@patch('blocks_genesis._core.api.AzureMessageClient')
@patch('blocks_genesis._core.api.MongoHandler')
async def test_close_lifespan(mock_mongo, mock_client):
    mock_client.get_instance.return_value.close = AsyncMock()
    mock_mongo._mongo_logger = MagicMock()
    await api.close_lifespan()
    mock_mongo._mongo_logger.stop.assert_called()

def test_configure_middlewares():
    app = FastAPI()
    api.configure_middlewares(app, is_local=True)
    from fastapi.testclient import TestClient
    with TestClient(app) as client:
        resp = client.get('/ping')
        assert resp.status_code == 200
        assert resp.json()['status'] == 'healthy' 


import pytest
import asyncio
from unittest.mock import AsyncMock, MagicMock, patch
from fastapi import HTTPException, Request
from blocks_genesis._auth import auth
from datetime import datetime, timezone

@pytest.mark.asyncio
@patch('aiohttp.ClientSession')
async def test_fetch_cert_bytes_http(mock_session):
    mock_resp = AsyncMock()
    mock_resp.read.return_value = b'certdata'
    mock_resp.raise_for_status.return_value = None
    mock_session.return_value.__aenter__.return_value = mock_session.return_value
    mock_session.return_value.get.return_value.__aenter__.return_value = mock_resp
    result = await auth.fetch_cert_bytes('http://example.com/cert')
    assert result == b'certdata'

@pytest.mark.asyncio
@patch('builtins.open', new_callable=MagicMock)
@patch('asyncio.get_running_loop')
async def test_fetch_cert_bytes_file(mock_loop, mock_open):
    mock_file = MagicMock()
    mock_file.read.return_value = b'certdata'
    mock_open.return_value.__enter__.return_value = mock_file
    mock_loop.return_value.run_in_executor = AsyncMock(return_value=b'certdata')
    result = await auth.fetch_cert_bytes('file.cert')
    assert result == b'certdata'

@pytest.mark.asyncio
@patch('blocks_genesis._auth.auth.fetch_cert_bytes', new_callable=AsyncMock)
async def test_get_tenant_cert_cache_miss(mock_fetch):
    cache_client = MagicMock()
    cache_client.get_string_value.return_value = None
    cache_client.add_string_value = AsyncMock()
    tenant = MagicMock()
    tenant.jwt_token_parameters.public_certificate_path = 'certpath'
    tenant.jwt_token_parameters.issue_date = datetime(2020, 1, 1, tzinfo=timezone.utc)
    tenant.jwt_token_parameters.certificate_valid_for_number_of_days = 10
    tenant.jwt_token_parameters.public_certificate_password = 'pass'
    mock_fetch.return_value = b'certdata'
    result = await auth.get_tenant_cert(cache_client, tenant, 'tid')
    assert result == b'certdata'
    cache_client.add_string_value.assert_awaited()

@patch('cryptography.hazmat.primitives.serialization.pkcs12.load_pkcs12')
def test_create_certificate_success(mock_load):
    cert = MagicMock()
    cert.additional_certs = [MagicMock(certificate='certobj')]
    mock_load.return_value = cert
    result = auth.create_certificate(b'data', 'pass')
    assert result == 'certobj'

@patch('cryptography.hazmat.primitives.serialization.pkcs12.load_pkcs12', side_effect=Exception('fail'))
def test_create_certificate_fail(mock_load):
    result = auth.create_certificate(b'data', 'pass')
    assert result is None

@pytest.mark.asyncio
@patch('blocks_genesis._auth.auth.get_tenant_cert', new_callable=AsyncMock)
@patch('blocks_genesis._auth.auth.create_certificate')
@patch('jwt.decode')
@patch('blocks_genesis._auth.auth.BlocksContextManager')
@patch('blocks_genesis._auth.auth.Activity')
async def test_authenticate_success(mock_activity, mock_context_mgr, mock_jwt_decode, mock_create_cert, mock_get_cert):
    request = MagicMock()
    request.headers.get.return_value = 'Bearer token'
    request.cookies.get.return_value = ''
    tenant_service = MagicMock()
    tenant = MagicMock()
    tenant.jwt_token_parameters.public_certificate_password = 'pass'
    tenant.jwt_token_parameters.issuer = 'issuer'
    tenant.jwt_token_parameters.audiences = 'aud'
    mock_context_mgr.get_context.return_value = MagicMock(tenant_id='tid')
    tenant_service.get_tenant = AsyncMock(return_value=tenant)
    mock_get_cert.return_value = b'certdata'
    cert = MagicMock()
    cert.public_key.return_value.public_bytes.return_value = b'keypem'
    mock_create_cert.return_value = cert
    mock_jwt_decode.return_value = {'sub': 'user'}
    mock_context_mgr.create_from_jwt_claims.return_value = MagicMock(user_id='user', roles=['role'], permissions=['perm'])
    result = await auth.authenticate(request, tenant_service, MagicMock())
    assert 'sub' in result

@pytest.mark.asyncio
async def test_authenticate_missing_token():
    request = MagicMock()
    request.headers.get.return_value = None
    request.cookies.get.return_value = ''
    with pytest.raises(HTTPException) as exc:
        await auth.authenticate(request, MagicMock(), MagicMock())
    assert exc.value.status_code == 401

@patch('blocks_genesis._auth.auth.get_tenant_service')
@patch('blocks_genesis._auth.auth.CacheProvider.get_client')
@patch('blocks_genesis._auth.auth.DbContext.get_provider')
@patch('blocks_genesis._auth.auth.authenticate', new_callable=AsyncMock)
@patch('blocks_genesis._auth.auth.BlocksContextManager.get_context')
def test_authorize_bypass(mock_get_context, mock_auth, mock_db, mock_cache, mock_tenant):
    mock_get_context.return_value = MagicMock(roles=['role'], permissions=['perm'], service_name='svc', tenant_id='tid')
    dep = auth.authorize(bypass_authorization=True)
    # Should return a Depends object
    assert dep is not None 


import pytest
from blocks_genesis._cache.CacheClient import CacheClient

class DummyCacheClient(CacheClient):
    def cache_database(self):
        raise NotImplementedError()
    def key_exists(self, key):
        raise NotImplementedError()
    def add_string_value(self, key, value, key_life_span=None):
        raise NotImplementedError()
    def get_string_value(self, key):
        raise NotImplementedError()
    def remove_key(self, key):
        raise NotImplementedError()
    def add_hash_value(self, key, value, key_life_span=None):
        raise NotImplementedError()
    def get_hash_value(self, key):
        raise NotImplementedError()
    async def key_exists_async(self, key):
        raise NotImplementedError()
    async def add_string_value_async(self, key, value, key_life_span=None):
        raise NotImplementedError()
    async def get_string_value_async(self, key):
        raise NotImplementedError()
    async def remove_key_async(self, key):
        raise NotImplementedError()
    async def add_hash_value_async(self, key, value, key_life_span=None):
        raise NotImplementedError()
    async def get_hash_value_async(self, key):
        raise NotImplementedError()
    async def publish_async(self, channel, message):
        raise NotImplementedError()
    async def subscribe_async(self, channel, handler):
        raise NotImplementedError()
    async def unsubscribe_async(self, channel):
        raise NotImplementedError()
    def dispose(self):
        raise NotImplementedError()
    async def dispose_async(self):
        raise NotImplementedError()

def test_context_manager():
    client = DummyCacheClient()
    with pytest.raises(NotImplementedError):
        with client:
            pass

import asyncio
@pytest.mark.asyncio
async def test_async_context_manager():
    client = DummyCacheClient()
    with pytest.raises(NotImplementedError):
        async with client:
            pass 


import pytest
from unittest.mock import AsyncMock, patch, MagicMock
from blocks_genesis._core.azure_key_vault import AzureKeyVault

@patch('blocks_genesis._core.azure_key_vault.EnvVaultConfig')
@patch('blocks_genesis._core.azure_key_vault.ClientSecretCredential')
@patch('blocks_genesis._core.azure_key_vault.SecretClient')
def test_azure_key_vault_init(mock_secret_client, mock_cred, mock_env):
    mock_env.get_config.return_value = {
        'KEYVAULT__CLIENTID': 'cid',
        'KEYVAULT__CLIENTSECRET': 'csec',
        'KEYVAULT__KEYVAULTURL': 'url',
        'KEYVAULT__TENANTID': 'tid',
    }
    vault = AzureKeyVault()
    assert vault.vault_url == 'url'
    assert mock_cred.called
    assert mock_secret_client.called

@pytest.mark.asyncio
@patch('blocks_genesis._core.azure_key_vault.AzureKeyVault.get_secret_value', new_callable=AsyncMock)
async def test_get_secrets(mock_get_secret_value):
    mock_get_secret_value.side_effect = lambda k: f'val-{k}'
    vault = AzureKeyVault.__new__(AzureKeyVault)
    vault.get_secret_value = mock_get_secret_value
    keys = ['A', 'B']
    result = await vault.get_secrets(keys)
    assert result == {'A': 'val-A', 'B': 'val-B'}

@pytest.mark.asyncio
@patch('blocks_genesis._core.azure_key_vault.SecretClient')
async def test_get_secret_value_success(mock_secret_client):
    vault = AzureKeyVault.__new__(AzureKeyVault)
    vault.secret_client = mock_secret_client.return_value
    mock_secret = MagicMock()
    mock_secret.value = 'v'
    vault.secret_client.get_secret = AsyncMock(return_value=mock_secret)
    result = await vault.get_secret_value('foo')
    assert result == 'v'

@pytest.mark.asyncio
@patch('blocks_genesis._core.azure_key_vault.SecretClient')
async def test_get_secret_value_error(mock_secret_client):
    vault = AzureKeyVault.__new__(AzureKeyVault)
    vault.secret_client = mock_secret_client.return_value
    vault.secret_client.get_secret = AsyncMock(side_effect=Exception('fail'))
    result = await vault.get_secret_value('foo')
    assert result == ''

@pytest.mark.asyncio
@patch('blocks_genesis._core.azure_key_vault.ClientSecretCredential')
@patch('blocks_genesis._core.azure_key_vault.SecretClient')
async def test_close(mock_secret_client, mock_cred):
    vault = AzureKeyVault.__new__(AzureKeyVault)
    vault.credential = MagicMock()
    vault.secret_client = MagicMock()
    vault.credential.close = AsyncMock()
    vault.secret_client.close = AsyncMock()
    await vault.close()
    vault.credential.close.assert_awaited()
    vault.secret_client.close.assert_awaited() 


import pytest
import os
import tempfile
import json
from blocks_genesis._core import configuration

def test_load_and_get_configurations(tmp_path, monkeypatch):
    config = {'foo': 'bar'}
    config_path = tmp_path / 'dev.json'
    with open(config_path, 'w') as f:
        json.dump(config, f)
    configuration.load_configurations(str(tmp_path))
    assert configuration.get_configurations()['foo'] == 'bar'

def test_load_configurations_file_not_found(tmp_path):
    with pytest.raises(FileNotFoundError):
        configuration.load_configurations(str(tmp_path))

def test_get_configurations_not_loaded():
    # Reset the module-level variable
    configuration._app_configurations = None
    with pytest.raises(Exception):
        configuration.get_configurations() 


from typing import Optional

from blocks_genesis._database.mongo_context import MongoDbContextProvider


class DbContext:
    _provider: Optional[MongoDbContextProvider] = None

    @classmethod
    def set_provider(cls, provider: MongoDbContextProvider) -> None:
        cls._provider = provider

    @classmethod
    def get_provider(cls) -> MongoDbContextProvider:
        if cls._provider is None:
            raise RuntimeError("No MongoDbContextProvider registered.")
        return cls._provider

    @classmethod
    def clear(cls) -> None:
        cls._provider = None



from azure.identity.aio import ClientSecretCredential
from azure.keyvault.secrets.aio import SecretClient
from typing import List, Dict
from blocks_genesis._core.env_vault_config import EnvVaultConfig


class AzureKeyVault:
    def __init__(self):
        required_keys = ["KEYVAULT__CLIENTID", "KEYVAULT__CLIENTSECRET", "KEYVAULT__KEYVAULTURL", "KEYVAULT__TENANTID"]
        config = EnvVaultConfig.get_config(required_keys)

        self.vault_url = config["KEYVAULT__KEYVAULTURL"]
        tenant_id = config["KEYVAULT__TENANTID"]
        client_id = config["KEYVAULT__CLIENTID"]
        client_secret = config["KEYVAULT__CLIENTSECRET"]

        self.credential = ClientSecretCredential(tenant_id, client_id, client_secret)
        self.secret_client = SecretClient(vault_url=self.vault_url, credential=self.credential)

    async def get_secrets(self, keys: List[str]) -> Dict[str, str]:
        secrets: Dict[str, str] = {}
        for key in keys:
            value = await self.get_secret_value(key)
            if value:
                secrets[key] = value
        return secrets

    async def get_secret_value(self, key: str) -> str:
        try:
            secret = await self.secret_client.get_secret(key)
            return secret.value
        except Exception as e:
            print(f"[Warning] Could not retrieve secret '{key}': {e}")
            return ""

    async def close(self):
        await self.credential.close()
        await self.secret_client.close()



import json
from typing import Any, Dict, List, Optional
from fastapi import Depends, Request, HTTPException
from pymongo.collection import Collection
from datetime import datetime, timezone
import base64
import aiohttp
import asyncio
import jwt
from jwt import ExpiredSignatureError, InvalidTokenError, PyJWKClient

from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.hazmat.primitives import serialization

from blocks_genesis._auth.blocks_context import BlocksContext, BlocksContextManager
from blocks_genesis._cache import CacheClient
from blocks_genesis._cache.cache_provider import CacheProvider
from blocks_genesis._database.db_context import DbContext
from blocks_genesis._lmt.activity import Activity
from blocks_genesis._tenant.tenant import Tenant
from blocks_genesis._tenant.tenant_service import TenantService, get_tenant_service



async def fetch_cert_bytes(cert_url: str) -> bytes:
    if cert_url.startswith("http"):
        async with aiohttp.ClientSession() as session:
            async with session.get(cert_url) as resp:
                resp.raise_for_status()
                return await resp.read()
    else:
        loop = asyncio.get_running_loop()
        try:
            with open(cert_url, "rb") as f:
                return await loop.run_in_executor(None, f.read)
        except Exception as e:
            raise RuntimeError(f"Error reading cert file {cert_url}: {e}")

async def get_tenant_cert(cache_client: CacheClient, tenant: Tenant, tenant_id: str) -> bytes:
    key = f"tetocertpublic::{tenant_id}"
    cert_bytes = cache_client.get_string_value(key)
    if cert_bytes is None:
        cert_bytes = await fetch_cert_bytes(tenant.jwt_token_parameters.public_certificate_path)
        now = datetime.now(timezone.utc)
        issue_date = tenant.jwt_token_parameters.issue_date
        if issue_date.tzinfo is None:
            issue_date = issue_date.replace(tzinfo=timezone.utc)
        days_remaining = (
            tenant.jwt_token_parameters.certificate_valid_for_number_of_days
            - (now - issue_date).days
            - 1
        )
        ttl = max(60, days_remaining  * 24 * 60 * 60)  # Ensure at least 60 seconds TTL
        if ttl > 0:
            cached_value = base64.b64encode(cert_bytes).decode("utf-8")
            await cache_client.add_string_value(key, cached_value, ex=int(ttl))
    return cert_bytes


async def authenticate(request: Request, tenant_service: TenantService, cache_client: CacheClient):

    tenant_id = BlocksContextManager.get_context().tenant_id if BlocksContextManager.get_context() else None
    tenant = await tenant_service.get_tenant(tenant_id)
    is_third_party_token = False

    header = request.headers.get("Authorization")
    if header and any(header.startswith(prefix) for prefix in ["bearer ", "Bearer "]):
        token = header.split(" ", 1)[1].strip()
    else:
        bc = BlocksContextManager.get_context()
        token = request.cookies.get(f"access_token_{bc.tenant_id}", "")
        if not token and (ck := getattr(tenant.third_party_jwt_token_parameters, "cookie_key", None)):
            token = request.cookies.get(ck, "")
            is_third_party_token = True
        
    if not token:
        raise HTTPException(401, "Token missing")
    
    if is_third_party_token:
     return await try_fallback_async(request=request, token=token, tenant = tenant , db_context=DbContext.get_provider())

    cert_bytes = await get_tenant_cert(cache_client, tenant, tenant_id)    
    cert = create_certificate(cert_bytes, tenant.jwt_token_parameters.public_certificate_password)
    if not cert:
        raise HTTPException(500, "Failed to load certificate")
    
    public_key = cert.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    public_key_str = public_key_pem.decode('utf-8')
    
    try:
        payload = jwt.decode(
            jwt=token,
            key=public_key_str,
            algorithms=["RS256"],
            issuer=tenant.jwt_token_parameters.issuer,
            audience=tenant.jwt_token_parameters.audiences,
            options={
                "verify_signature": True,  
                "verify_exp": True,        
                "verify_iss": True,        
                "verify_aud": True, 
                "verify_iat": True,
                "verify_nbf": True,      
                "require":["exp", "iat", "iss", "aud", "nbf"]
            },
            leeway=0 
        )
        extended_payload = dict(payload)
        extended_payload[BlocksContext.REQUEST_URI_CLAIM] = str(request.url)
        extended_payload[BlocksContext.TOKEN_CLAIM] = token

        blocks_context = BlocksContextManager.create_from_jwt_claims(extended_payload)
        BlocksContextManager.set_context(blocks_context)
        Activity.set_current_property("baggage.UserId", blocks_context.user_id)
        Activity.set_current_property("baggage.IsAuthenticate", "true")
        
        return extended_payload
    except ExpiredSignatureError as e:
      print(f"JWT expired: {e}")
      raise HTTPException(401, "Token expired")
    
    except InvalidTokenError as e:
     print(f"JWT verification failed: {e}, trying fallback...")

     validation_result = await try_fallback_async(request=request, token=token,tenant=tenant, db_context=DbContext.get_provider())

     if not validation_result:
        raise HTTPException(401, f"Invalid token: {e}")

     return validation_result

    

def create_certificate(certificate_data: bytes, password: str = None):
    try:
        password_bytes = password.encode('utf-8') if password else None
        certificate = pkcs12.load_pkcs12(certificate_data, password_bytes)
        return certificate.additional_certs[0].certificate
    except Exception as e:
        print(f"Failed to create certificate: {e}")
        return None
    

async def try_fallback_async(
    request,
    token: str,
    tenant: Tenant,
    db_context: DbContext,
    ex: Optional[Exception] = None
) -> bool:
  
    if ex is not None:
        print(f"[Fallback] Triggered due to: {type(ex).__name__} - {ex}")

    try:
        if not token or token.strip() == "":
            print("[Fallback] ❌ No token found in request.")
            return False

        tp = getattr(tenant, "third_party_jwt_token_parameters", None)
        if tp is None:
            print("[Fallback] ❌ Tenant has no ThirdPartyJwtTokenParameters.")
            return False

        # Prefer JWKS if present
        if tp.jwks_url:
            payload = await _get_from_jwks_url(token, tp)
        else:
            payload = await _get_from_public_certificate(token, tp)

        if payload is None:
            print("[Fallback] ❌ Fallback validation failed.")
            return False

        validated = await _validate_token_with_fallback_async(payload, token, request, db_context)
        return validated

    except Exception as final_ex:
        print(f"[Fallback] 💥 Unhandled error: {final_ex}")
        return False


# --------------------------
# JWKS-based retrieval + validation
# --------------------------
async def _get_from_jwks_url(token: str, tp_params) -> Optional[Dict[str, Any]]:

    jwks_url = tp_params.jwks_url
    if not jwks_url:
        return None

    try:
        # PyJWKClient is synchronous; run in threadpool to avoid blocking event loop
        loop = asyncio.get_running_loop()

        def _get_signing_key():
            client = PyJWKClient(jwks_url)
            signing_key = client.get_signing_key_from_jwt(token)
            return signing_key.key

        public_key = await loop.run_in_executor(None, _get_signing_key)

        # decode
        options = {"verify_signature": True, "verify_exp": True, "verify_nbf": True, "verify_iat": True}
        payload = jwt.decode(
            token,
            key=public_key,
            algorithms=["RS256", "RS384", "RS512"],
            issuer=tp_params.issuer or None,
            audience=tp_params.audiences or None,
            options=options,
            leeway=0
        )
        print("[Fallback] ✅ Token validated via JWKS.")
        return payload

    except Exception as e:
        print(f"[Fallback] ❌ JWKS validation failed: {e}")
        return None


# --------------------------
# Public certificate-based retrieval + validation
# --------------------------
async def _get_from_public_certificate(token: str, tp_params) -> Optional[Dict[str, Any]]:
   
    try:
        cert_path = tp_params.public_certificate_path
        if not cert_path:
            print("[Fallback] ❌ No public certificate path configured.")
            return None

        cert_bytes = await fetch_cert_bytes(cert_path)
        if cert_bytes is None:
            print("[Fallback] ❌ Failed to load public certificate bytes.")
            return None

        cert = create_certificate(cert_bytes, tp_params.public_certificate_password or "")
        if cert is None:
            print("[Fallback] ❌ Failed to create certificate object.")
            return None

        public_key = cert.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        options = {"verify_signature": True, "verify_exp": True, "verify_nbf": True, "verify_iat": True}
        payload = jwt.decode(
            token,
            key=public_key,
            algorithms=["RS256", "RS384", "RS512"],
            issuer=tp_params.issuer or None,
            audience=tp_params.audiences or None,
            options=options,
            leeway=0
        )

        print("[Fallback] ✅ Token validated via public certificate.")
        return payload

    except Exception as e:
        print(f"[Fallback] ❌ Public certificate validation failed: {e}")
        return None


# --------------------------
# Final validation step + store BlocksContext
# --------------------------
async def _validate_token_with_fallback_async(
    payload: Dict[str, Any],
    token: str,
    request,
    db_context: DbContext
) -> bool:
    
    try:
        # Attach request uri and token to payload (like AddClaims in C#)
        payload = dict(payload)  # make a copy
        payload[BlocksContext.REQUEST_URI_CLAIM] = str(request.url)
        payload[BlocksContext.TOKEN_CLAIM] = token

        # Map and store third-party context fields (reads mapping from DB)
        await _store_third_party_blocks_context_activity(payload, request, db_context)
        # Activity baggage
        Activity.set_current_property("baggage.UserId", BlocksContextManager.get_context().user_id or "")
        Activity.set_current_property("baggage.IsAuthenticate", "true")

        print("[Fallback] ✅ Fallback flow finished successfully.")
        return True

    except Exception as e:
        print(f"[Fallback] ❌ Validation+context store failed: {e}")
        return False


# --------------------------
# Claim mapping & BlocksContext creation (C# StoreThirdPartyBlocksContextActivity equivalent)
# --------------------------
async def _store_third_party_blocks_context_activity(
    payload: Dict[str, Any],
    request,
    db_context: DbContext
):
    try:
        # find mapping doc (the C# code takes the first doc)
        coll = await  DbContext.get_provider().get_collection("ThirdPartyJWTClaims")
        claims_mapper =  coll.find_one({}) or {}

        # helper functions to parse nested claim object like "user.roles" or "roles"
        def _extract_claim_property(claim_object: str) -> str:
            parts = claim_object.split(".")
            return parts[-1] if parts else claim_object

        def _get_claim_object_name(claim_object: str) -> str:
            parts = claim_object.split(".")
            return parts[0] if parts else claim_object

        def _extract_claim_value_from_payload(payload_obj: Dict[str, Any], claim_object: str) -> str:
            # If claim_object is nested like "profile.name" - traverse
            parts = claim_object.split(".")
            cur = payload_obj
            try:
                for p in parts:
                    if isinstance(cur, dict):
                        cur = cur.get(p)
                    else:
                        return ""
                # if the final is not a primitive, return its JSON string representation
                if isinstance(cur, (dict, list)):
                    return json.dumps(cur)
                return str(cur) if cur is not None else ""
            except Exception:
                return ""

        #roles resolution
        roles: List[str] = []
        role_claims_key = claims_mapper.get("Roles", "")

        if role_claims_key:
            obj_name = _get_claim_object_name(role_claims_key)
            claim_value = payload.get(obj_name)
            roles = claim_value[_extract_claim_property(role_claims_key)]

        # fallback: if still empty, look for standard role claim type
        if not roles:
            # common role claim keys
            for k in (role_claims_key, "http://schemas.microsoft.com/ws/2008/06/identity/claims/role"):
                roleClaims = payload.get(k)
                if(not roleClaims):
                    roles = roleClaims
                    break

        # user identifier
        sub_claim = payload.get("sub", "") or payload.get("nameid", "") or payload.get("user_id", "")
        email_claim = payload.get("email", "")

        # Build userId according to mapping (C# logic)
        user_id_mapper = claims_mapper.get("UserId", "")
        if user_id_mapper:
            # if mapping says "sub" then keep sub, otherwise extract nested value and append "_external"
            if _extract_claim_property(user_id_mapper) == "sub":
                user_id = sub_claim
            else:
                extracted = _extract_claim_value_from_payload(payload, user_id_mapper)
                user_id = (extracted + "_external") if extracted else (sub_claim + "_external" if sub_claim else "")
        else:
            user_id = sub_claim or ""

        # username and display name mapping
        def _map_field(mapper_key: str, fallback_field: str = "") -> str:
            mapper_val = claims_mapper.get(mapper_key, "")
            if mapper_val.lower() == "email":
                return email_claim or ""
            if mapper_val:
                return _extract_claim_value_from_payload(payload, mapper_val)
            return fallback_field or ""

        user_name = _map_field("UserName", "")
        display_name = _map_field("Name", "")

        # email mapping: if not set, try mapper
        email = email_claim or _extract_claim_value_from_payload(payload, claims_mapper.get("Email", "") or "")

        api_key = request.headers.get("x-blocks-key", "")  # C# used header BlocksConstants.BlocksKey

        # Create BlocksContext (you might have a factory method; adapt as needed)
        blocks_ctx = BlocksContextManager.create(
            tenant_id=api_key,
            roles= roles,
            user_id=user_id,
            is_authenticated=True,
            request_uri=str(request.url),
            organization_id="",
            email=email or "",
            permissions=[],
            user_name=user_name or "",
            phone_number="",
            display_name=display_name or "",
            oauth_token=payload.get("oauth", None),
            actual_tenant_id=api_key
        )

        BlocksContextManager.set_context(blocks_ctx)

        # Store serialized context on request.state so downstream handlers can access it.
        request.state.third_party_context_json = json.dumps(blocks_ctx.dict() if hasattr(blocks_ctx, "dict") else {})

        # also optionally set a header-like bag in request.state for downstream response middleware to copy into a header
        request.state.third_party_context_header = request.state.third_party_context_json

        print("[Fallback] ✅ Third-party context stored from claims mapper.")
    except Exception as e:
        print(f"[Fallback] ❌ Error storing third-party blocks context: {e}")



async def extract_project_key(request: Request) -> str | None:
    # 1. Query param
    if "ProjectKey" in request.query_params:
        v = request.query_params.get("ProjectKey")
        if v:
            return v

    # 2. JSON body (must allow buffering)
    try:
        body_bytes = await request.body()
        if body_bytes:
            body_json = json.loads(body_bytes.decode("utf-8"))
            project_key = body_json.get("projectKey")
            if project_key:
                return project_key
    except Exception:
        pass

    return None


async def is_project_owner_or_shared(user_id: str, project_key: str, db_context: DbContext, tenants: TenantService):
    # Check if this tenant is created by the user (owner check)
    tenant = await tenants.get_tenant(project_key)
    if tenant and tenant.created_by == user_id:
        return True

    # Check shared project (ProjectPeoples collection)
    collection = await db_context.get_collection("ProjectPeoples", tenant_id=project_key)
    query = {"UserId": user_id, "TenantId": project_key}
    shared = await collection.find_one(query)
    return shared is not None


async def handle_root_tenant_access(
    request: Request,
    context: BlocksContext,
    tenants: TenantService,
    db_context: DbContext
) -> bool:

    blocks_key = request.headers.get("blocks-key")
    if not blocks_key:
        return False

    tenant = await tenants.get_tenant(blocks_key)
    if tenant is None or not tenant.is_root_tenant:
        return False

    project_key = await extract_project_key(request)
    if not project_key:
        return False

    user_id = context.user_id
    if not user_id:
        return False

    allowed = await is_project_owner_or_shared(user_id, project_key, db_context, tenants)
    return allowed


async def check_standard_access(
    context: BlocksContext,
    controller: str,
    action: str,
    db_context: DbContext
) -> bool:

    roles = context.roles or []
    permissions = context.permissions or []

    resource = f"{context.service_name}::{controller}::{action}".lower()

    collection = await db_context.get_collection("Permissions", tenant_id=context.tenant_id)

    query = {
        "Type": 1,
        "Resource": resource,
        "$or": [
            {"Roles": {"$in": roles}},
            {"Name": {"$in": permissions}}
        ]
    }

    count = await collection.count_documents(query)
    return count > 0


def authorize(bypass_authorization: bool = False):
    async def dependency(request: Request):
        tenant_service = TenantService()
        cache_client = CacheProvider.get_client()
        db_context = DbContext.get_provider()

        # 1. Authenticate (your JWT logic)
        await authenticate(request, tenant_service, cache_client)

        context = BlocksContextManager.get_context()
        if not context:
            raise HTTPException(401, "Missing context")

        # 2. Bypass if requested
        if bypass_authorization:
            return

        # 3. Extract controller & action (same as C# ControllerActionDescriptor)
        path_parts = request.url.path.strip("/").split("/")
        if len(path_parts) >= 4:
            controller = path_parts[2]
            action = path_parts[3]
        elif len(path_parts) >= 2:
            controller = path_parts[0]
            action = path_parts[1]
        else:
            raise HTTPException(400, "Invalid URL format.")

        # ---------------------------------------------------------
        #         ROOT TENANT SPECIAL ACCESS (C# equivalent)
        # ---------------------------------------------------------
        is_root_allowed = await handle_root_tenant_access(
            request=request,
            context=context,
            tenants=tenant_service,
            db_context=db_context
        )

        if is_root_allowed:
            return  # Allowed immediately like context.Succeed()

        # ---------------------------------------------------------
        #              STANDARD PERMISSION CHECK
        # ---------------------------------------------------------
        allowed = await check_standard_access(
            context=context,
            controller=controller,
            action=action,
            db_context=db_context
        )

        if not allowed:
            raise HTTPException(403, "Insufficient permissions")

    return Depends(dependency)



from datetime import datetime
from typing import List, Optional
from pydantic import BaseModel, Field
from enum import IntEnum

from blocks_genesis._entities.base_entity import BaseEntity


# ------------------------------
# Certificate Storage Enum
# ------------------------------
class CertificateStorageType(IntEnum):
    AZURE = 1
    FILESYSTEM = 2
    MONGODB = 3


# ------------------------------
# JWT Token Parameters (MAIN)
# ------------------------------
class JwtTokenParameters(BaseModel):
    issuer: Optional[str] = Field(alias="Issuer", default="")
    subject: Optional[str] = Field(alias="Subject", default="")
    audiences: List[str] = Field(default_factory=list, alias="Audiences")
    public_certificate_path: Optional[str] = Field(alias="PublicCertificatePath", default="")
    public_certificate_password: Optional[str] = Field(alias="PublicCertificatePassword", default="")
    private_certificate_password: Optional[str] = Field(alias="PrivateCertificatePassword", default="")
    certificate_storage_type: CertificateStorageType = Field(
        alias="CertificateStorageType",
        default=CertificateStorageType.AZURE
    )
    certificate_valid_for_number_of_days: int = Field(
        alias="CertificateValidForNumberOfDays",
        default=365
    )
    issue_date: Optional[datetime] = Field(alias="IssueDate", default=None)

    class Config:
        extra = "ignore"
        validate_by_name = True
        use_enum_values = True


# ------------------------------
# Third Party JWT Token Parameters
# ------------------------------
class ThirdPartyJwtTokenParameters(BaseModel):
    provider_name: str = Field(alias="ProviderName", default="")
    issuer: str = Field(alias="Issuer", default="")
    subject: str = Field(alias="Subject", default="")
    audiences: List[str] = Field(alias="Audiences", default_factory=list)
    public_certificate_path: str = Field(alias="PublicCertificatePath", default="")
    jwks_url: str = Field(alias="JwksUrl", default="")
    public_certificate_password: str = Field(alias="PublicCertificatePassword", default="")
    cookie_key: str = Field(alias="CookieKey", default="")

    class Config:
        extra = "ignore"
        validate_by_name = True


# ------------------------------
# Tenant Entity
# ------------------------------
class Tenant(BaseEntity):
    tenant_id: Optional[str] = Field(alias="TenantId", default="")
    is_accept_blocks_terms: bool = Field(alias="IsAcceptBlocksTerms", default=False)
    is_use_blocks_exclusively: bool = Field(alias="IsUseBlocksExclusively", default=False)

    name: Optional[str] = Field(alias="Name", default=None)
    db_name: str = Field(alias="DBName", default="")
    application_domain: str = Field(alias="ApplicationDomain", default="")
    allowed_domains: List[str] = Field(alias="AllowedDomains", default_factory=list)
    cookie_domain: str = Field(alias="CookieDomain", default="")

    is_disabled: bool = Field(alias="IsDisabled", default=False)
    db_connection_string: str = Field(alias="DbConnectionString", default="")
    tenant_salt: str = Field(alias="TenantSalt", default="")

    jwt_token_parameters: Optional[JwtTokenParameters] = Field(alias="JwtTokenParameters", default=None)
    third_party_jwt_token_parameters: Optional[ThirdPartyJwtTokenParameters] = Field(
        alias="ThirdPartyJwtTokenParameters",
        default=None
    )

    is_root_tenant: bool = Field(alias="IsRootTenant", default=False)
    is_domain_verified: bool = Field(alias="IsDomainVerified", default=False)

    environment: Optional[str] = Field(alias="Environment", default="")
    tenant_group_id: Optional[str] = Field(alias="TenantGroupId", default="")
    custom_domain: Optional[str] = Field(alias="CustomDomain", default="")

    class Config:
        extra = "ignore"
        validate_by_name = True



import asyncio
from datetime import datetime, timedelta
from typing import Dict, Optional, Callable, Any
import redis
import redis.asyncio as aioredis
from opentelemetry.trace import StatusCode

from blocks_genesis._auth.blocks_context import BlocksContextManager
from blocks_genesis._cache.CacheClient import CacheClient
from blocks_genesis._core.secret_loader import get_blocks_secret
from blocks_genesis._lmt.activity import Activity


class RedisClient(CacheClient):
    """Redis client implementation with Activity tracing"""
    
    def __init__(self):
        """Initialize Redis client"""
        self._blocks_secret = get_blocks_secret()
        self._connection_string = self._blocks_secret.CacheConnectionString
        self._subscriptions: Dict[str, Callable] = {}
        self._disposed = False
        self._pubsub_tasks: Dict[str, asyncio.Task] = {}
        
        # Parse connection string and initialize clients
        self._redis_config = self._parse_connection_string(self._blocks_secret.CacheConnectionString)
        self._sync_client = redis.Redis(**self._redis_config)
        self._async_client: Optional[aioredis.Redis] = None
    
    def _parse_connection_string(self, connection_string: str) -> Dict[str, Any]:
        parts = connection_string.split(',', 1)
        host_port = parts[0]  # e.g. "hostname:6379"
        
        query = ""
        if len(parts) > 1:
            query = parts[1].replace(',', '&')
        
        url = f"redis://{host_port}"
        if query:
            url += f"/?{query}"
        
        config = redis.connection.parse_url(url)

        # Allowed keys for redis.Redis constructor
        allowed_keys = {
            'host', 'port', 'username', 'password', 'db', 'ssl', 
            'socket_connect_timeout', 'socket_timeout',
            'encoding', 'encoding_errors', 'decode_responses',
            'retry_on_timeout', 'max_connections'
        }

        # Map keys like connectTimeout -> socket_connect_timeout (in seconds)
        if 'connectTimeout' in config:
            # convert from ms to seconds
            config['socket_connect_timeout'] = int(config.pop('connectTimeout')) / 1000
        if 'syncTimeout' in config:
            config['socket_timeout'] = int(config.pop('syncTimeout')) / 1000

        # Remove unsupported keys
        config = {k: v for k, v in config.items() if k in allowed_keys}

        return config
 
    async def _get_async_client(self) -> aioredis.Redis:
        """Get or create async Redis client"""
        if self._async_client is None:
            self._async_client = aioredis.Redis(**self._redis_config)
        return self._async_client
    
    def cache_database(self) -> redis.Redis:
        """Get the Redis database instance"""
        return self._sync_client
    
    def _create_activity(self, key: str, operation: str) -> Activity:
        """Create activity for tracing"""
        activity = Activity.start(f"Redis::{operation}")
        activity.set_property("key", key)
        activity.set_property("operation", operation)
        activity.set_property("baggage.TenantId", BlocksContextManager.get_context().tenant_id if BlocksContextManager.get_context() else "miscellaneous")
        return activity
    
    # Synchronous Methods
    def key_exists(self, key: str) -> bool:
        """Check if key exists"""
        with self._create_activity(key, "KeyExists") as activity:
            try:
                result = self._sync_client.exists(key) > 0
                activity.set_property("exists", result)
                return result
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    def add_string_value(self, key: str, value: str, key_life_span: Optional[int] = None) -> bool:
        """Add string value to cache"""
        with self._create_activity(key, "AddStringValue") as activity:
            try:
                activity.set_property("value_length", len(value))
                if key_life_span is not None:
                    activity.set_property("ttl", key_life_span)
                    result = self._sync_client.setex(key, key_life_span, value)
                else:
                    result = self._sync_client.set(key, value)
                
                success = bool(result)
                activity.set_property("success", success)
                return success
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    def get_string_value(self, key: str) -> Optional[str]:
        """Get string value from cache"""
        with self._create_activity(key, "GetStringValue") as activity:
            try:
                result = self._sync_client.get(key)
                activity.set_property("found", result is not None)
                if result:
                    activity.set_property("value_length", len(result))
                return result
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    def remove_key(self, key: str) -> bool:
        """Remove key from cache"""
        with self._create_activity(key, "RemoveKey") as activity:
            try:
                result = self._sync_client.delete(key) > 0
                activity.set_property("deleted", result)
                return result
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    def add_hash_value(self, key: str, value: Dict[str, Any], key_life_span: Optional[int] = None) -> bool:
        """Add hash value to cache"""
        with self._create_activity(key, "AddHashValue") as activity:
            try:
                activity.set_property("field_count", len(value))
                if key_life_span is not None:
                    activity.set_property("ttl", key_life_span)
                
                self._sync_client.hset(key, mapping=value)
                if key_life_span is not None:
                    success = bool(self._sync_client.expire(key, key_life_span))
                else:
                    success = True
                
                activity.set_property("success", success)
                return success
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    def get_hash_value(self, key: str) -> Dict[str, Any]:
        """Get hash value from cache"""
        with self._create_activity(key, "GetHashValue") as activity:
            try:
                result = self._sync_client.hgetall(key)
                hash_dict = dict(result) if result else {}
                activity.set_property("found", bool(result))
                activity.set_property("field_count", len(hash_dict))
                return hash_dict
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    # Asynchronous Methods
    async def key_exists_async(self, key: str) -> bool:
        """Check if key exists (async)"""
        client = await self._get_async_client()
        with self._create_activity(key, "KeyExists") as activity:
            try:
                result = await client.exists(key) > 0
                activity.set_property("exists", result)
                return result
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    async def add_string_value_async(self, key: str, value: str, key_life_span: Optional[int] = None) -> bool:
        """Add string value to cache (async)"""
        client = await self._get_async_client()
        with self._create_activity(key, "AddStringValue") as activity:
            try:
                activity.set_property("value_length", len(value))
                if key_life_span is not None:
                    activity.set_property("ttl", key_life_span)
                    result = await client.setex(key, key_life_span, value)
                else:
                    result = await client.set(key, value)
                
                success = bool(result)
                activity.set_property("success", success)
                return success
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    async def get_string_value_async(self, key: str) -> Optional[str]:
        """Get string value from cache (async)"""
        client = await self._get_async_client()
        with self._create_activity(key, "GetStringValue") as activity:
            try:
                result = await client.get(key)
                activity.set_property("found", result is not None)
                if result:
                    activity.set_property("value_length", len(result))
                return result
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    async def remove_key_async(self, key: str) -> bool:
        """Remove key from cache (async)"""
        client = await self._get_async_client()
        with self._create_activity(key, "RemoveKey") as activity:
            try:
                result = await client.delete(key) > 0
                activity.set_property("deleted", result)
                return result
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    async def add_hash_value_async(self, key: str, value: Dict[str, Any], key_life_span: Optional[int] = None) -> bool:
        """Add hash value to cache (async)"""
        client = await self._get_async_client()
        with self._create_activity(key, "AddHashValue") as activity:
            try:
                activity.set_property("field_count", len(value))
                if key_life_span is not None:
                    activity.set_property("ttl", key_life_span)
                
                await client.hset(key, mapping=value)
                if key_life_span is not None:
                    success = bool(await client.expire(key, key_life_span))
                else:
                    success = True
                
                activity.set_property("success", success)
                return success
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    async def get_hash_value_async(self, key: str) -> Dict[str, Any]:
        """Get hash value from cache (async)"""
        client = await self._get_async_client()
        with self._create_activity(key, "GetHashValue") as activity:
            try:
                result = await client.hgetall(key)
                hash_dict = dict(result) if result else {}
                activity.set_property("found", bool(result))
                activity.set_property("field_count", len(hash_dict))
                return hash_dict
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    # Pub/Sub Methods
    async def publish_async(self, channel: str, message: str) -> int:
        """Publish message to channel"""
        if not channel:
            raise ValueError("Channel cannot be empty")
        
        client = await self._get_async_client()
        with self._create_activity(channel, "Publish") as activity:
            try:
                activity.set_property("message_length", len(message))
                result = await client.publish(channel, message)
                activity.set_property("subscribers_notified", result)
                return result
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    async def subscribe_async(self, channel: str, handler: Callable[[str, str], None]) -> None:
        """Subscribe to channel with handler"""
        if not channel:
            raise ValueError("Channel cannot be empty")
        if handler is None:
            raise ValueError("Handler cannot be None")
        
        client = await self._get_async_client()
        with self._create_activity(channel, "Subscribe") as activity:
            try:
                # Store the handler
                self._subscriptions[channel] = handler
                
                # Create pubsub instance
                pubsub = client.pubsub()
                await pubsub.subscribe(channel)
                
                # Create task to handle messages
                task = asyncio.create_task(self._handle_subscription(pubsub, channel, handler))
                self._pubsub_tasks[channel] = task
                
                activity.set_property("subscribed", True)
                
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                self._subscriptions.pop(channel, None)
                raise
    
    async def unsubscribe_async(self, channel: str) -> None:
        """Unsubscribe from channel"""
        if not channel:
            raise ValueError("Channel cannot be empty")
        
        with self._create_activity(channel, "Unsubscribe") as activity:
            try:
                # Cancel the subscription task
                if channel in self._pubsub_tasks:
                    task = self._pubsub_tasks.pop(channel)
                    task.cancel()
                    try:
                        await task
                    except asyncio.CancelledError:
                        pass
                
                # Remove from subscriptions
                self._subscriptions.pop(channel, None)
                activity.set_property("unsubscribed", True)
                
            except Exception as ex:
                activity.set_property("error", True)
                activity.set_property("error_message", str(ex))
                activity.set_status(StatusCode.ERROR, str(ex))
                raise
    
    async def _handle_subscription(self, pubsub: aioredis.client.PubSub, channel: str, handler: Callable[[str, str], None]):
        """Handle subscription messages"""
        try:
            async for message in pubsub.listen():
                if message['type'] == 'message':
                    with Activity.start("Redis::MessageReceived") as message_activity:
                        message_activity.set_property("channel", channel)
                        try:
                            channel_name = message['channel']
                            if isinstance(channel_name, bytes):
                                channel_name = channel_name.decode('utf-8')
                            
                            message_data = message['data']
                            if isinstance(message_data, bytes):
                                message_data = message_data.decode('utf-8')
                            
                            message_activity.set_property("message_length", len(message_data))
                            handler(channel_name, message_data)
                            message_activity.set_property("handled", True)
                            
                        except Exception as ex:
                            message_activity.set_property("error", True)
                            message_activity.set_property("error_message", str(ex))
                            message_activity.set_status(StatusCode.ERROR, str(ex))
                            self._logger.error(f"Error handling message in channel {channel}: {ex}")
        except asyncio.CancelledError:
            # Expected when unsubscribing
            pass
        except Exception as ex:
            self._logger.error(f"Error in subscription handler for channel {channel}: {ex}")
        finally:
            await pubsub.unsubscribe(channel)
            await pubsub.close()
    
    # Dispose pattern
    def dispose(self) -> None:
        """Dispose resources synchronously"""
        if self._disposed:
            return
        
        # Clean up subscriptions
        for channel in list(self._subscriptions.keys()):
            self._subscriptions.pop(channel, None)
        
        # Close sync client
        if self._sync_client:
            self._sync_client.close()
        
        self._disposed = True
    
    async def dispose_async(self) -> None:
        """Dispose resources asynchronously"""
        if self._disposed:
            return
        
        # Cancel all pubsub tasks
        for channel, task in self._pubsub_tasks.items():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
        
        self._pubsub_tasks.clear()
        self._subscriptions.clear()
        
        # Close async client
        if self._async_client:
            await self._async_client.close()
        
        # Close sync client
        if self._sync_client:
            self._sync_client.close()
        
        self._disposed = True
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dispose()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.dispose_async()


import pytest
import asyncio
from unittest.mock import AsyncMock, MagicMock, patch
from blocks_genesis._core.worker import WorkerConsoleApp

@pytest.mark.asyncio
@patch('blocks_genesis._core.worker.SecretLoader')
@patch('blocks_genesis._core.worker.configure_logger')
@patch('blocks_genesis._core.worker.configure_tracing')
@patch('blocks_genesis._core.worker.CacheProvider')
@patch('blocks_genesis._core.worker.RedisClient')
@patch('blocks_genesis._core.worker.initialize_tenant_service', new_callable=AsyncMock)
@patch('blocks_genesis._core.worker.DbContext')
@patch('blocks_genesis._core.worker.MongoDbContextProvider')
@patch('blocks_genesis._core.worker.EventRegistry')
@patch('blocks_genesis._core.worker.get_blocks_secret')
@patch('blocks_genesis._core.worker.ConfigAzureServiceBus')
@patch('blocks_genesis._core.worker.AzureMessageClient')
@patch('blocks_genesis._core.worker.AzureMessageWorker')
@patch('blocks_genesis._core.worker.MessageConfiguration')
async def test_setup_services_and_cleanup(
    mock_msg_config, mock_worker, mock_client, mock_config, mock_get_secret, mock_event_registry,
    mock_mongo_provider, mock_dbcontext, mock_init_tenant, mock_redis, mock_cache, mock_tracing, mock_logger, mock_secret_loader
):
    # Setup
    msg_config = MagicMock()
    msg_config.connection = None
    mock_msg_config.return_value = msg_config
    mock_get_secret.return_value.MessageConnectionString = 'connstr'
    mock_event_registry._handlers = {}
    mock_event_registry.register = lambda event_type: lambda handler: handler
    mock_worker_instance = MagicMock()
    mock_worker.return_value = mock_worker_instance
    mock_worker_instance.initialize.return_value = None
    mock_worker_instance.stop = AsyncMock()
    mock_secret_loader.return_value.load_secrets = AsyncMock()
    app = WorkerConsoleApp('test', msg_config, {'evt': lambda: None})
    # Test setup_services
    async with app.setup_services() as worker:
        assert worker is mock_worker_instance
    # Test cleanup
    await app.cleanup()

@pytest.mark.asyncio
@patch('blocks_genesis._core.worker.SecretLoader')
@patch('blocks_genesis._core.worker.configure_logger')
@patch('blocks_genesis._core.worker.configure_tracing')
@patch('blocks_genesis._core.worker.CacheProvider')
@patch('blocks_genesis._core.worker.RedisClient')
@patch('blocks_genesis._core.worker.initialize_tenant_service', new_callable=AsyncMock)
@patch('blocks_genesis._core.worker.DbContext')
@patch('blocks_genesis._core.worker.MongoDbContextProvider')
@patch('blocks_genesis._core.worker.EventRegistry')
@patch('blocks_genesis._core.worker.get_blocks_secret')
@patch('blocks_genesis._core.worker.ConfigAzureServiceBus')
@patch('blocks_genesis._core.worker.AzureMessageClient')
@patch('blocks_genesis._core.worker.AzureMessageWorker')
@patch('blocks_genesis._core.worker.MessageConfiguration')
async def test_setup_services_event_registration_error(
    mock_msg_config, mock_worker, mock_client, mock_config, mock_get_secret, mock_event_registry,
    mock_mongo_provider, mock_dbcontext, mock_init_tenant, mock_redis, mock_cache, mock_tracing, mock_logger, mock_secret_loader
):
    msg_config = MagicMock()
    msg_config.connection = None
    mock_msg_config.return_value = msg_config
    mock_get_secret.return_value.MessageConnectionString = 'connstr'
    mock_event_registry._handlers = {'evt': lambda: None}
    mock_event_registry.register = lambda event_type: lambda handler: handler
    mock_worker_instance = MagicMock()
    mock_worker.return_value = mock_worker_instance
    mock_worker_instance.initialize.return_value = None
    mock_worker_instance.stop = AsyncMock()
    mock_secret_loader.return_value.load_secrets = AsyncMock()
    app = WorkerConsoleApp('test', msg_config, {'evt': lambda: None})
    async with app.setup_services() as worker:
        assert worker is mock_worker_instance
    await app.cleanup()

@pytest.mark.asyncio
@patch('blocks_genesis._core.worker.WorkerConsoleApp.setup_services')
async def test_run(mock_setup_services):
    app = WorkerConsoleApp('test', MagicMock())
    mock_worker = MagicMock()
    mock_worker.run = AsyncMock()
    mock_setup_services.return_value.__aenter__.return_value = mock_worker
    await app.run()
    mock_worker.run.assert_awaited()

@pytest.mark.asyncio
@patch('blocks_genesis._core.worker.WorkerConsoleApp.setup_services')
async def test_run_keyboard_interrupt(mock_setup_services):
    app = WorkerConsoleApp('test', MagicMock())
    mock_worker = MagicMock()
    async def raise_interrupt():
        raise KeyboardInterrupt()
    mock_worker.run = raise_interrupt
    mock_setup_services.return_value.__aenter__.return_value = mock_worker
    await app.run() 


import asyncio
from collections.abc import Callable
from contextlib import asynccontextmanager
import logging
from typing import Any, Dict, Type, Union

from blocks_genesis._cache.cache_provider import CacheProvider
from blocks_genesis._cache.redis_client import RedisClient
from blocks_genesis._core.secret_loader import SecretLoader, get_blocks_secret
from blocks_genesis._database.db_context import DbContext
from blocks_genesis._database.mongo_context import MongoDbContextProvider
from blocks_genesis._message.azure.azure_message_client import AzureMessageClient
from blocks_genesis._message.azure.azure_message_worker import AzureMessageWorker
from blocks_genesis._message.azure.config_azure_service_bus import ConfigAzureServiceBus
from blocks_genesis._message.event_registry import EventRegistry
from blocks_genesis._message.message_configuration import MessageConfiguration
from blocks_genesis._lmt.log_config import configure_logger
from blocks_genesis._lmt.mongo_log_exporter import MongoHandler
from blocks_genesis._lmt.tracing import configure_tracing
from blocks_genesis._tenant.tenant_service import initialize_tenant_service


class WorkerConsoleApp:
    def __init__(self, name: str, message_config: MessageConfiguration, register_consumer: Dict[str, Union[Callable[..., Any], Type[Any]]] = None):
        """
        Initializes the WorkerConsoleApp.

        Args:
            name (str): The name of the worker application.
            message_config (MessageConfiguration): Configuration for message bus connectivity.
            register_consumer (Dict[str, Union[Callable[..., Any], Type[Any]]], optional):
                A dictionary mapping event types (strings) to their corresponding handlers.
                Handlers can be callables or classes with a 'handle' method.
                Defaults to an empty dictionary if not provided.
        """
        self.message_worker: AzureMessageWorker = None
        self.name = name
        self.logger = logging.getLogger(__name__)
        self.message_config = message_config
        self.register_consumer = register_consumer if register_consumer is not None else {}

    @asynccontextmanager
    async def setup_services(self):
        """
        An asynchronous context manager to set up all necessary services for the worker.
        It handles initialization of secrets, logging, tracing, caching, database context,
        message bus configuration, and registers event handlers.
        """
        self.logger.info("🚀 Starting Blocks AI Worker Console App...")

        try:
            self.logger.info("Loading secrets...")
            await SecretLoader(self.name).load_secrets()
            self.logger.info("Secrets loaded successfully")
            
            configure_logger()
            self.logger.info("Logger configured")

            configure_tracing()
            self.logger.info("Tracing configured")

            CacheProvider.set_client(RedisClient())
            await initialize_tenant_service()
            DbContext.set_provider(MongoDbContextProvider())
            self.logger.info("Cache, TenantService, and Mongo Context initialized")

            for event_type, handler in self.register_consumer.items():
                if isinstance(event_type, str) and event_type and (callable(handler) or hasattr(handler, "handle")):
                    if event_type not in EventRegistry._handlers:
                        EventRegistry.register(event_type)(handler)
                        self.logger.info(f"Handler registered for event type: {event_type}")
                    else:
                        self.logger.error(f"Handler already registered for event type: {event_type}")
                else:
                    self.logger.error(f"Invalid event_type or handler for: {event_type} (Expected non-empty string event_type and callable/handle-method handler).")

            
            self.message_config.connection = self.message_config.connection or get_blocks_secret().MessageConnectionString
            ConfigAzureServiceBus().configure_queue_and_topic(self.message_config)
            AzureMessageClient.initialize(self.message_config)

            self.message_worker = AzureMessageWorker(self.message_config)
            self.message_worker.initialize()

            self.logger.info("Azure Message Worker initialized and ready")
            yield self.message_worker

        except Exception as ex:
            self.logger.error(f"Startup failed: {ex}", exc_info=True)
            raise 

        finally:
            await self.cleanup()

    async def cleanup(self):
        """
        Performs asynchronous cleanup operations before the application exits.
        This includes stopping the message worker and potentially the Mongo logger.
        """
        self.logger.info("Cleaning up services...")

        if self.message_worker:
            self.logger.info("Stopping Azure Message Worker...")
            await self.message_worker.stop()
            self.logger.info("Azure Message Worker stopped.")

        if hasattr(MongoHandler, '_mongo_logger') and MongoHandler._mongo_logger:
            self.logger.info("Stopping Mongo log exporter...")
            MongoHandler._mongo_logger.stop()
            self.logger.info("Mongo log exporter stopped.")

        self.logger.info("✅ Shutdown complete")

    async def run(self, callback: Union[Callable[..., Any], Type[Any]]):
        """
        Runs the worker application using the setup_services context manager.
        It handles graceful shutdown on cancellation or keyboard interrupt.
        """
        async with self.setup_services() as worker:
            self.logger.info("Worker running... Press Ctrl+C to stop")
            if callback and callable(callback):
                await callback()
            try:
                await worker.run() 
            except asyncio.CancelledError:
                self.logger.info("Received cancellation signal (e.g., from task.cancel()).")
            except KeyboardInterrupt:
                self.logger.info("Received interrupt signal (Ctrl+C). Initiating graceful shutdown.")
            finally:
                self.logger.info("Worker run loop exited.")




import logging

from blocks_genesis._lmt.mongo_log_exporter import MongoHandler, TraceContextFilter


def configure_logger():
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter(
        "[%(asctime)s] [%(levelname)s] [%(TenantId)s] [%(TraceId)s] %(message)s"
    )

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    
    mongo_handler = MongoHandler()

    # Add context filter to enrich log records
    context_filter = TraceContextFilter()
    console_handler.addFilter(context_filter)
    mongo_handler.addFilter(context_filter)

    logger.handlers.clear()
    logger.addHandler(console_handler)
    logger.addHandler(mongo_handler)
